local D = LibStub:NewLibrary("DoomCore-1.3",1)
if not D then return end
local SharedMedia = LibStub( "LibSharedMedia-3.0" )
DoomCore = D
tconcat = tconcat or table.concat


-----------
-- Utility
-----------

local function bound(minBound, val, maxBound)
  if minBound and val < minBound then return minBound
  elseif maxBound and val > maxBound then return maxBound
  else return val
  end
end
D.bound = bound

local function empty(val)
  return val == nil or val == "" or (type(val) == "table" and next(val) == nil)
end
D.empty = empty

local function say(...)
  local arguments = { ... }
  local output = {}
  for i=1,#arguments do
    output[i] = tostring(arguments[i])
  end
  return tconcat(output,"")
end
D.say = say
local function tprint(table) for k,v in pairs(table) do print(k,v) end end
D.tprint = tprint

local function tstartswith(from, match)
  for i=1,#match do
    if from[i] ~= match[i] then return false end
  end
  return true
end
D.tstartswith = tstartswith

local function stringify(val)
  if type(val) == "table" then
    if empty(val) then return "{}"
    elseif #val > 0 then
      local output = {}
      for i=1,#val do
        output[i] = tostring(val[i])
      end
      return "{" .. tconcat(output, ", ") .. "}"
    else
      local output = {}
      for k,v in pairs(val) do
        tinsert(output, tostring(k) .. " = " .. tostring(v))
      end
      return "{" .. tconcat(output, ", ") .. "}"
    end
  end
  return tostring(val)
end
D.stringify = stringify

local function assertNotNil(...)
  local arguments = { ... }
  local failures = {}
  for i=1,#arguments do
    if arguments[i] == nil then
      tinsert(failures, say("Arg #",ceil(i/2)," is nil."))
    end
  end
  if #failures>0 then error(tconcat(failures," "),2) end
  return ...
end
D.assertNotNil = assertNotNil

local function assertType(...)
  local arguments = { ... }
  local failures = {}
  for i=1,#arguments-1,2 do
    local isType = type(arguments[i])
    local matchType = { strsplit("/", arguments[i+1]) }
    local match = false
    for _,subType in pairs(matchType) do
      if isType == subType then
        match = true
        break
      end
    end
    if not match then
      tinsert(failures, say("Arg #",ceil(i/2)," must be ",arguments[i+1],", is ",isType, "."))
    end
  end
  if #failures>0 then error(tconcat(failures," "),tonumber(arguments[#arguments]) or 2) end
end
D.assertType = assertType

local function nilSort(val, valFunc, sortFunc)
  local reverse = false
  local valFuncType = type(valFunc)
  if valFuncType == "string" and valFunc:sub(1,1) == "-" then
    reverse = true
    if #valFunc < 2 then valFunc = nil else valFunc = valFunc:sub(2) end
  end
  sort(val, function(a, b)
    if valFuncType == "string" then
      if a ~= nil then a = a[valFunc] end
      if b ~= nil then b = b[valFunc] end
    elseif valFuncType == "function" then
      if a ~= nil then a = valFunc(a) end
      if b ~= nil then b = valFunc(b) end
    end
    if a == nil and b == nil then
      return false
    end
    if a == nil then
      return false
    end
    if b == nil then
      return true
    end
    if sortFunc then return sortFunc(a,b)
    elseif reverse then return b < a
    else return a < b
    end
  end)
end
D.nilSort = nilSort

local function capitalize(val)
  assertType(val,"string", 3)
  return val:sub(1, 1):upper() .. val:sub(2):lower()
end
D.capitalize = capitalize

local function reverse(l)
  local newL = {}
  for i = getn(l), 1, -1 do tinsert(newL, l[i]) end
  return newL
end
D.reverse = reverse

local function flip(flip, ...)
  if flip then
    return unpack(reverse{ ... }) else
    return ...
  end
end
D.flip = flip

local function roundString(val)
  assertType(val,"number", 3)
  if val > 1000000 then
    return floor(val / 10000) / 100 .. "m"
  else
    return floor(val / 1000) .. "k"
  end
end
D.roundString = roundString

local function orderNum(text)
  local math_pow = math.pow
  assertType(text,"string", 3)
  text = text:upper()
  local order = 0
  for i=1,#text do
    order = order + (strbyte(text, i)) / math_pow(100, i)
  end
  return order
end
D.orderNum = orderNum


local function map(func, before, subArgs, after)
  assertType(func,"function", before,"table/nil", subArgs,"table/nil", after,"table/nil", 3)
  local r = {}
  local result
  local args = { unpack(before or {}) }
  local slot = #args + 1
  if after then for i=1,#after do args[slot + i] = after[i] end end
  for i=1,#subArgs do
    local arg = subArgs[i]
    args[slot] = arg
    result = { func(unpack(args)) }
    if empty(result) then result = nil elseif #result == 1 then result = result[1] end
    r[i] = result
  end
  return unpack(r)
end
D.map = map

local function hexToCol(code)
  if empty(code) then return { r = 1, g = 1, b = 1, a = 1 } end
  return {
    r = tonumber("0x"..code:sub(1,2)) / 255,
    g = tonumber("0x"..code:sub(3,4)) / 255,
    b = tonumber("0x"..code:sub(5,6)) / 255,
    a = 1
  }
end
D.hexToCol = hexToCol

local function colUnpack(col)
  if #col >= 3 then return unpack(col) end
  return col.r, col.g, col.b, col.a
end
D.colUnpack = colUnpack

local function colToHue(col)
  assertType(col,"table", 3)
  local col_r, col_g, col_b = colUnpack(col)
  local hue
  local minCol = min(col_r, col_g, col_b)
  local maxCol = max(col_r, col_g, col_b)
  local range = maxCol - minCol
  if maxCol == col_r then
    hue = (col_g - col_b) / range
  elseif maxCol == col_g then
    hue = 2 + (col_b - col_r) / range
  else
    hue = 4 + (col_r - col_g) / range
  end
  hue = hue * 60
  if hue < 0 then hue = hue + 360 end
  return hue
end
D.colToHue = colToHue

local function getItemID(val)
  local itemString = select(2, GetItemInfo(val))
  if type(itemString) == "string" then
    local subString = itemString:sub(itemString:find(":") + 1)
    subString = subString:sub(1, subString:find(":") - 1)
    return tonumber(subString)
  end
end
D.getItemID = getItemID


-----------
-- Anchors
-----------

local anchors = {
  center      = "Center",
  topleft     = "Top Left",
  top         = "Top",
  topright    = "Top Right",
  left        = "Left",
  right       = "Right",
  bottomleft  = "Bottom Left",
  bottom      = "Bottom",
  bottomright = "Bottom Right"
}
D.anchors = anchors

local VERTICAL = 1
D.VERTICAL = VERTICAL
local HORIZONTAL = 2
D.HORIZONTAL = HORIZONTAL
local orientations = {
  horizontal = "Horizontal",
  vertical = "Vertical"
}
D.orientations = orientations
local orientGrowth = {
  horizontal = {
    [false] = "right",
    [true] = "left"
  },
  vertical = {
    [false] = "top",
    [true] = "bottom"
  }
}
D.orientGrowth = orientGrowth
local direction = {
  TOP    = VERTICAL,    top    = VERTICAL,
  BOTTOM = VERTICAL,    bottom = VERTICAL,
  LEFT   = HORIZONTAL,  left   = HORIZONTAL,
  RIGHT  = HORIZONTAL,  right  = HORIZONTAL,
  horizontal = HORIZONTAL,
  vertical = VERTICAL
}
D.direction = direction
local function compound(anchor1, anchor2)
  local dir1, dir2 = direction[anchor1], direction[anchor2]
  if not dir1 then return anchor1 end
  if not dir2 then return anchor2 end
  if dir1 < dir2 then
    return anchor1 .. anchor2
  else
    return anchor2 .. anchor1
  end
end
D.compound = compound

local grow = { left = "Left", right = "Right", top = "Up", bottom = "Down" }
D.grow = grow

local growAnchors = {
  left   = { "RIGHT",  "LEFT",  -1,   0 },
  right  = { "LEFT",   "RIGHT",  1,   0 },
  top    = { "BOTTOM", "TOP",    0,   1 },
  bottom = { "TOP",    "BOTTOM", 0,  -1 },
  blank  = { "CENTER", "CENTER", 0,   0 },
  center = { "CENTER", "CENTER", 0,   0 }
}
D.growAnchors = growAnchors
for _,dir1 in pairs{ "left", "right" } do for _,dir2 in pairs{ "top", "bottom" } do
  growAnchors[compound(growAnchors[dir1][1],growAnchors[dir2][1])] = { nil, nil, growAnchors[dir1][3], growAnchors[dir2][4] }
end end

local corners = {
  topleft     = { "BOTTOMRIGHT", "TOPLEFT",     -1, -1 },
  topright    = { "BOTTOMLEFT",  "TOPRIGHT",     1, -1 },
  bottomleft  = { "TOPRIGHT",    "BOTTOMLEFT",  -1,  1 },
  bottomright = { "TOPLEFT",     "BOTTOMRIGHT",  1,  1 }
}
D.corners = corners

local borders = {
  top    = { { "BOTTOMLEFT", "TOPLEFT",    -1,  0 }, { "BOTTOMRIGHT", "TOPRIGHT",     1,  0 },  0,  1 },
  bottom = { { "TOPLEFT",    "BOTTOMLEFT", -1,  0 }, { "TOPRIGHT",    "BOTTOMRIGHT",  1,  0 },  0,  1 },
  left   = { { "TOPRIGHT",   "TOPLEFT",     0,  1 }, { "BOTTOMRIGHT", "BOTTOMLEFT",   0, -1 },  1,  0 },
  right  = { { "TOPLEFT",    "TOPRIGHT",    0,  1 }, { "BOTTOMLEFT",  "BOTTOMRIGHT",  0, -1 },  1,  0 }
}
D.borders = borders

local function convertDims(dir, dim1, dim2)
  if direction[dir] == HORIZONTAL then
    return dim1, dim2
  else
    return dim2, dim1
  end
end
D.convertDims = convertDims


----------
-- Frames
----------

local function fitButton(button,dW,dH)
  local width,height = button:GetFontString():GetSize()
  button:SetSize(width + (dW or 0), height + (dH or 0))
end
D.fitButton = fitButton

local function enable(frame) frame:EnableMouse(true) end
D.enable = enable

local function Locker()
  local locker = DoomCoreLocker
  if locker then return locker end
  locker = CreateFrame("Button", "DoomCoreLocker", UIParent, "UIPanelButtonTemplate")
  locker:SetFrameStrata("DIALOG")
  locker.frames = {}
  locker:SetText("Lock all")
  locker:SetSize(80, 25)
  locker:SetPoint("BOTTOM", 0, 150)
  locker:Hide()
  function locker:Update()
    local visible = false
    for frameName, frame in pairs(self.frames) do
      if frame and frame.conf and frame.conf.lock == false then
        visible = true
        break
      end
    end
    if visible then self:Show() else self:Hide() end
  end
  locker:SetScript("OnClick", function(self)
    for frameName, frame in pairs(self.frames) do
      if frame.conf then frame.conf.lock = true end
      if frame.tex then frame.tex:Hide() end
    end
    self.frames = {}
    self:Hide()
  end)
  return locker
end
D.Locker = Locker

local function updateFrame(frame, conf, anchor, parent)
  conf = conf or frame.conf
  anchor = anchor or conf.anchor
  parent = parent or UIFrame
  frame:ClearAllPoints()
  local unlocked = false
  if conf then
    if conf.width then frame:SetWidth(conf.width) end
    if conf.height then frame:SetHeight(conf.height) end
    if conf.anchor and conf.offsetX and conf.offsetY then
      frame:SetPoint(anchor, parent, conf.anchor, conf.offsetX, conf.offsetY)
    end
    if conf.lock ~= nil then
      unlocked = conf.lock
      local locker = Locker()
      locker.frames[frame:GetName()] = frame
      locker:Update()
    end
  end

  local unlocked = conf and conf.lock == false
  frame:SetMovable(unlocked)
  frame:EnableMouse(unlocked)
  if not frame.tex then
  elseif unlocked then
    frame.tex:Show()
  else
    frame.tex:Hide()
  end
  for _,child in pairs{ frame:GetChildren() } do
    if child.EnableMouse then child:EnableMouse(not unlocked) end
  end
  return unlocked
end
D.updateFrame = updateFrame

local function tooltip(button, show, conf, onHide)
  conf = conf or {}
  button:SetMotionScriptsWhileDisabled(show)
  button:SetScript("OnEnter", function(button)
    if not button:IsShown() then return end
    if conf.tooltipOverride then
      GameTooltip:SetOwner(button, "ANCHOR_" .. conf.tooltipAnchor)
    else
      GameTooltip:SetOwner(button)
    end
    local buttonType = button:GetAttribute("type")
    local subject = buttonType and button:GetAttribute(buttonType)
    if button.tooltip then
      GameTooltip:SetText(button.tooltip)
    elseif buttonType == "spell" then
      if type(subject) ~= "number" then
        subject = select(7, GetSpellInfo(subject))
      end
      GameTooltip:SetSpellByID(subject)
    elseif buttonType == "item" then
      GameTooltip:SetItemByID(subject)
    end
    GameTooltip:Show()
  end)
  button:SetScript("OnLeave", function(button)
    if not button:IsMouseOver() then
      GameTooltip:FadeOut()
    end
  end)
  button:SetScript("OnHide", function(button)
    if button:IsMouseOver() then
      GameTooltip:FadeOut()
    end
    if onHide then onHide(button) end
  end)
end
D.tooltip = tooltip

local function draggable(frame, frameSettings, func)
  if frame:GetScript("OnReceiveDrag") then return end
  frame:RegisterForDrag("LeftButton")
  frame:SetScript("OnDragStart", frame.StartMoving)
  frame:SetScript("OnDragStop", frame.StopMovingOrSizing)
  frame:SetScript("OnReceiveDrag", function()
    local orig,_,dest,x,y = frame:GetPoint(1)
    local a = growAnchors[orig]
    if frameSettings then
      frameSettings.anchor = dest:lower()
      frameSettings.offsetX = floor(x*10)/10
      frameSettings.offsetY =  floor(y*10)/10
    end
    if func then func() end
  end)
end
D.draggable = draggable

local function setBg(frame, background, border, edge, inset)
  edge = edge or 4
  inset = inset or 4
  local bg = {}
  if background and background ~= "None" then
    bg.bgFile = SharedMedia:Fetch(SharedMedia.MediaType.BACKGROUND, background)
  end
  if border and border ~= "None" then
    bg.edgeFile = SharedMedia:Fetch(SharedMedia.MediaType.BORDER, border)
    bg.edgeSize = edge
  end
  bg.insets = { left = inset, right = inset, top = inset, bottom = inset }
  frame:SetBackdrop(bg)
end
D.setBg = setBg

local function makeGrid(conf)
  conf.padding = conf.padding or 0
  conf.spacing = conf.spacing or 0

  if not conf.grow2 then
    if direction[conf.grow] == HORIZONTAL then
      conf.grow2 = conf.rowGrowth
    else
      conf.grow2 = conf.columnGrowth
    end
  end

  local conf_frame, conf_parent = conf.frame, conf.parent
  local conf_els, conf_max, conf_limit = conf.els, conf.max, conf.limit
  local conf_anchor, conf_grow, conf_grow2 = conf.anchor, conf.grow, conf.grow2
  local conf_size, conf_spacing, conf_padding = conf.size, conf.spacing, conf.padding
  local conf_x, conf_y = conf.x, conf.y
  local conf_background, conf_border, conf_edge = conf.background, conf.border, conf.edge
  local conf_inset = conf.inset

  local a1 = growAnchors[conf_grow] or growAnchors.blank
  local a2 = growAnchors[conf_grow2] or growAnchors.blank
  local comp = compound(a1[1], a2[1])
  local growComp = growAnchors[comp]

  conf_frame:ClearAllPoints()
  conf_frame:SetPoint(comp, conf_parent, conf_anchor,
    --conf_x - (conf_size + conf_padding * 2) * growComp[3] / 2,
    --conf_y - (conf_size + conf_padding * 2) * growComp[4] / 2)
    conf_x, conf_y)

  if not conf_els or #conf_els == 0 then
    conf_frame:SetSize(conf_size + conf_padding * 2, conf_size + conf_padding * 2)
    return
  end

  conf_els[1]:ClearAllPoints()
  conf_els[1]:SetPoint(comp, conf_frame, comp, growComp[3] * conf_padding, growComp[4] * conf_padding)
  conf_els[1]:Show()

  local max = #conf_els
  if conf_max and conf_max > 0 then max = min(max, conf_max) end

  for i=2,max do
    conf_els[i]:ClearAllPoints()

    if conf_limit and conf_limit > 0 and (i-1) % conf_limit == 0 then
      conf_els[i]:SetPoint(a2[1], conf_els[i-conf_limit], a2[2], a2[3] * conf_spacing, a2[4] * conf_spacing)
    else
      conf_els[i]:SetPoint(a1[1], conf_els[i-1], a1[2], a1[3] * conf_spacing, a1[4] * conf_spacing)
    end
    conf_els[i]:Show()
  end
  for i=max+1,#conf_els do
    conf_els[i]:Hide()
  end
  local dim1, dim2
  if conf_limit and conf_limit > 0 and max > conf_limit then
    dim1 = conf_padding * 2 + conf_size * conf_limit + conf_spacing * (conf_limit - 1)
    local subs = ceil(max/conf_limit)
    dim2 = conf_padding * 2 + conf_size * subs + conf_spacing * (subs - 1)
  else
    dim1 = conf_padding * 2 + conf_size * max + conf_spacing * (max - 1)
    dim2 = conf_padding * 2 + conf_size
  end
  conf_frame:SetSize(convertDims(conf_grow, dim1, dim2))
  setBg(conf_frame, conf_background, conf_border, conf_edge, conf_inset)
  --conf_frame:Show()
end
D.makeGrid = makeGrid

local ScrollGrid = function(name, secure, conf, parent, masque)
  parent = parent or UIParent
  local inherit = "ActionButtonTemplate"
  if secure then inherit = inherit .. ",SecureActionButtonTemplate" end
  local frame = CreateFrame("Frame", name, parent)
  local slider = CreateFrame("Slider", name .. ".Slider", frame)
  slider:SetPoint("TOPLEFT", frame, "TOPRIGHT", -6, -10)
  slider:SetPoint("BOTTOMLEFT", frame, "BOTTOMRIGHT", -6, 12)
  slider:SetThumbTexture("Interface\\DialogFrame\\UI-DialogBox-Background-Dark")
  slider:GetThumbTexture():SetSize(5,15)
  slider:SetWidth(5)
  frame.Slider = slider
  frame:EnableMouseWheel(true)
  frame:SetScript("OnMouseWheel", function(self, delta)
    self.i = bound(0, (self.i or 0) + self.conf.cols * delta * -1, self.maxI or 0)
    self.Slider:SetValue(self.i)
    self:Display()
  end)

  if masque then
    frame.msq = masque:Group(name)
  end
  frame.conf = conf
  local inner = CreateFrame("Frame", name .. ".Inner", frame)
  inner:SetAllPoints()
  inner:Show()
  frame.Inner = inner

  function frame:FindMaxI()
    self.maxI = 0
    local els, rows, cols = self.els, self.conf.rows, self.conf.cols
    if not els then return end
    local elsSize = #els
    if elsSize == 0 then return end
    local remainder = elsSize % cols
    local full = elsSize
    if remainder > 0 then full = full + (cols - remainder) end
    self.maxI = max(0, full - cols * rows)
    self.Slider:SetMinMaxValues(0, self.maxI)
    self.Slider:SetValue(self.i)
  end

  function frame:Refresh(show)
    if InCombatLockdown() then return print("Can't modify",self:GetName(),"in combat.") end
    local conf,inner = self.conf, self.Inner
    local padding, cols, rows = conf.padding, conf.cols, conf.rows
    local iconSize, spacing = conf.iconSize, conf.spacing
    local charmSize = conf.charmSize or 14
    self:SetSize(
      padding * 2 + cols * iconSize + (cols - 1) * spacing,
      padding * 2 + rows * iconSize + (rows - 1) * spacing
    )
    local displayCount = cols * rows
    local children = { inner:GetChildren() }
    local childrenCount = #children

    local display = {}

    for i = 1, displayCount do
      local child = children[i]
      if not child then
        child = CreateFrame("Button", self:GetName() .. i, inner, inherit)
        child.Charms = {}
        child.Overlays = {}
        if self.msq then self.msq:AddButton(child) end
        children[i] = child
      end
      for _,v in pairs(child.Charms) do v:SetSize(charmSize, charmSize) end
      for _,v in pairs(child.Overlays) do v:SetSize(charmSize, charmSize) end
      tooltip(child, conf.tooltip, conf)
      child:SetSize(iconSize, iconSize)
      tinsert(display, child)
    end
    for i = displayCount + 1, childrenCount do
      children[i]:Hide()
    end

    makeGrid{
      frame        = inner,
      els          = display,
      parent       = self,
      size         = iconSize,
      padding      = padding,
      spacing      = spacing,
      limit        = cols,
      grow         = "right",
      grow2        = "bottom",
    }
    self.display = display

    updateFrame(self, conf)
    setBg(self, conf.background, conf.border, conf.edge, conf.inset)
    self:FindMaxI()
    local slider = self.Slider
    self:Display(show)
  end

  function frame:SetEls(els)
    self.els = els
    self.i = 0
    self:FindMaxI()
  end

  function frame:GetButtons()
    return self.Inner:GetChildren()
  end

  function frame:Display(show)
    if InCombatLockdown() then
      if show == false then
        print("Can't hide",self:GetName(),"in combat.")
      else
        print("Can't display",self:GetName(),"in combat.")
      end
      return
    end
    if show == false then return self:Hide() end
    local i, display, els, conf = self.i or 0, self.display, self.els, self.conf
    local charmSize = conf.charmSize or 14
    if not els then return end
    for childI,child in pairs(display) do
      local el = els[i + childI]
      if not el then
        child:Hide()
      else
        for _,charm in pairs(child.Charms) do charm:Hide() end
        for _,charm in pairs(child.Overlays) do charm:Hide() end
        --child:SetAttribute("unit","player")
        for k,v in pairs(el) do
          if k == "icon" then
            child.icon:SetTexture(v)
          elseif k == "tooltip" then
            child.tooltip = v
          elseif k == "item" or k == "spell" or k == "macro" then
            child:SetAttribute("type", k)
            if k == "macro" then k = "macrotext" end
            child:SetAttribute(k, v)
          elseif k == "charms" and v then
            local i = 1
            local charms = child.Charms
            for _,tex in pairs(v) do
              if tex then
                local charm = charms[i]
                if not charm then
                  charm = child:CreateTexture(child:GetName() .. ".Charms[" .. i .. "]", "OVERLAY")
                  charm:SetDrawLayer("OVERLAY", min(i, 7))
                  if i == 1 then
                    charm:SetPoint("TOPLEFT", child, "TOPLEFT", -1, 3)
                  else
                    charm:SetPoint("LEFT", charms[i-1], "RIGHT", 0.5, 0)
                  end
                  charm:SetSize(charmSize, charmSize)
                  charms[i] = charm
                end
                if tex == "star" then
                  charm:SetTexture(922035)
                  charm:SetTexCoord(0.13,0.72,0.08,0.67)
                else
                  charm:SetTexture(tex)
                  charm:SetTexCoord(0,1,0,1)
                end
                charm:Show()
                i = i + 1
              end
            end
          elseif k == "overlays" and v then
            local i = 1
            local overlays = child.Overlays
            for iLevel, over in pairs(v) do
              if over then
                local overlay = overlays[i]
                if not overlay then
                  overlay = child:CreateTexture(child:GetName() .. ".Overlays[" .. i .. "]", "OVERLAY")
                end
                local tex, color, anchor = over.tex, over.color, { unpack(over.anchor) }
                local size = over.size or charmSize
                tinsert(anchor, 2, child)
                overlay:ClearAllPoints()
                overlay:SetPoint(unpack(anchor))
                overlay:SetSize(size, size)
                overlays[i] = overlay
                if tex then
                  if tex == "star" then
                    overlay:SetTexture(922035)
                    overlay:SetTexCoord(0.13,0.72,0.08,0.67)
                  else
                    overlay:SetTexture(tex)
                    overlay:SetTexCoord(0,1,0,1)
                  end
                  if color then overlay:SetVertexColor(colUnpack(color))
                  else overlay:SetVertexColor(1,1,1)
                  end
                elseif color then
                  overlay:SetColorTexture(colUnpack(color))
                end
                overlay:SetDrawLayer("OVERLAY",-7 + min(iLevel, 7))
                overlay:Show()
                i = i + 1
              end
            end
          else
            child:SetAttribute(k,v)
          end
        end
        child:Show()
      end
    end
    if self.msq then self.msq:ReSkin() end
    if show then self:Show() end
  end

  frame:Refresh()
  return frame
end
D.ScrollGrid = ScrollGrid

-------------------
-- Object-oriented
-------------------

local function setIndex(obj, ind, noOverride)
  local meta = getmetatable(obj)
  if meta == nil then
    meta = {}
    setmetatable(obj, meta)
  end
  if noOverride and meta.__index then return end
  meta.__index = ind
end
D.setIndex = setIndex

local Inherits = {}
function Inherits:New()
  self.__index = self
  local new = {}
  setmetatable(new, self)
  new._super = self
  return new
end

function Inherits:Map(funcName, before, subArgs, after)
  before = { self, unpack(before or {}) }
  map(self[funcName], before, subArgs, after)
end


------------
-- Database
------------

local Key = Inherits:New()
Key:New()
function Key:New(copyVal)
  copyVal = copyVal or {}
  assertType(copyVal,"table", 3)
  local val = { unpack(copyVal) }
  setmetatable(val, self)
  return val
end
function Key:s(size) -- Sub-array
  assertType(size,"number", 3)
  local newVal = Key:New{}
  for i=1,size do
    newVal[i] = self[i]
  end
  return newVal
end
function Key:r(remove) -- Remove
  assertType(remove,"number", 3)
  return self:s(#self-remove)
end
function Key:a(...) -- Append
  local newVal = Key:New(self)
  for _,v in pairs{...} do
    tinsert(newVal, v)
  end
  return newVal
end
D.Key = Key

local function K(val)
  return Key:New(val)
end
D.K = K

local function npairs(val, depth, ...)
  if val == nil or type(val) ~= "table" then return pairs{} end
  local i = {}
  local done = false

  if depth <= 1 then return function()
    if done then return end
    done = true
    return val
  end end

  local subPairs = npairs(val, depth - 1, select(2, ...))
  local result = {}
  local check = ...

  local function assignNext()
    local subVal = result[depth - 1]
    if subVal == nil or type(subVal) ~= "table" then
      i[depth -1], i[depth] = nil, nil
      return
    end
    while true do
      i[depth -1], i[depth] = next(subVal, i[depth -1])
      if i[depth -1]==nil then return end
      local checkType = type(check)
      if not check and i[depth -1]~=nil then return end
      if checkType == "function" and check(i[depth]) then return end
    end
  end

  return function()
    assignNext()
    while not done and i[depth] == nil do
      result = { subPairs() }
      if result[1] == nil then
        done = true
        return
      else
        i = Key:New(result):r(1)
        assignNext()
      end
    end
    if not done then return unpack(i) end
  end
end
D.npairs = npairs


-------------------
-- Crawling
-------------------

local function crawl(obj, key, noChildren, raw)
  assertType(obj,"table", key,"table", 3)
  local max = #key
  if max == 0 then return obj elseif max < 0 then return nil end
  local node = obj
  local child
  for i=1,max do
    if raw then
      child = rawget(node, key[i])
    else
      child = node[key[i]]
    end
    if not noChildren and child == nil then
      child = {}
      node[key[i]] = child
    end
    node = child
    if node == nil then return end
  end
  return node
end
D.crawl = crawl

local NodeCrawler = Inherits:New()
function NodeCrawler:New(core)
  core = core or {}
  assertType(core,"table", 3)
  local new = NodeCrawler._super.New(self)
  new.core = core
  return new
end
function NodeCrawler:Get(key)
  assertType(key,"table", 3)
  return crawl(self.core, key, true)
end
function NodeCrawler:Set(key, value, noOverride)
  assertType(key,"table", 3)
  if empty(key) then
    if value == nil then value = {} end
    if not noOverride then self.core = value end
    return self.core
  end
  local parent = crawl(self.core, K(key):r(1), false, true)
  if parent ~= nil and type(parent) ~= "table" then return end
  local child = key[#key]
  if not (noOverride and parent[child] ~= nil) then
    parent[child] = value
  end
  return parent[child]
end
function NodeCrawler:GetOrMake(key)
  assertType(key,"table", 3)
  return self:Set(key, {}, true)
end
function NodeCrawler:Incr(key, amount)
  amount = amount or 1
  assertType(key,"table", amount,"number", 3)
  local parent = crawl(self.core, K(key):r(1), false, true)
  local child = key[#key]
  parent[child] = amount + (parent[child] or 0)
  return parent[child]
end
function NodeCrawler:Append(key, ...)
  assertType(key,"table", 3)
  local parent = crawl(self.core, K(key):r(1), false, true)
  if parent ~= nil and type(parent) ~= "table" then return end
  local childKey = key[#key]
  local child = parent[childKey]
  if child == nil then
    child = {}
    parent[childKey] = child
  elseif type(child) ~= "table" then
    return
  end
  for _,el in pairs{...} do
    tinsert(parent[childKey], el)
  end
end
D.NodeCrawler = NodeCrawler

local Configuration = NodeCrawler:New()
function Configuration:New(val)
  local new = Configuration._super.New(self, val)
  return new
end
function Configuration:AutoDefault(child)
  if not child or not child.type or not self.defaults then return end
  local default = self.defaults[child.type]
  if default then setIndex(child, default, true) end
end

function Configuration:CrawlDefaults(node, crawled)
  crawled = crawled or {}
  if not node then node = self.core end
  crawled[node] = true
  if type(node) ~= "table" then return end
  self:AutoDefault(node)
  for k,v in pairs(node) do
    if not crawled[v] then
      if type(v) == "table" and not (type(k) == "string" and k:find("_")) then
        self:CrawlDefaults(v, crawled)
      end
    end
  end
end
function Configuration:Make(parent, makeType, childName, noOverride)
  assertType(parent,"table", makeType,"string", childName,"string/number", 3)
  if type(childName) == "string" then childName = childName:gsub("_","") end
  if empty(childName) then return end
  if noOverride and Configuration:Overriden(parent, childName) then return parent[childName] end
  local child = parent[childName]
  if child == nil then
    child = {}
    parent[childName] = child
  end
  child.type = makeType
  self:AutoDefault(child)
  return child
end
function Configuration:GetAll(parent, getType, depth)
  if empty(parent) then return pairs{} end
  depth = depth or 2
  assertType(parent,"table", getType,"string", depth,"number", 3)
  return npairs(parent, depth,
    function(val)
      return type(val) == "table" and val.type == getType
    end
  )
end
function Configuration:Count(parent, getType, depth)
  assertType(parent,"table", getType,"string", 3)
  local count = 0
  for ele in self:GetAll(parent, getType, depth) do
    count = count+1
  end
  return count
end
function Configuration:Delete(parent, childName)
  assertType(parent,"table", childName,"string/number", 3)
  parent[childName] = nil
  parent._exclude = parent._exclude or {}
  parent._exclude[childName] = true
end
function Configuration:Excluded(parent, childName)
  assertType(parent,"table", childName,"string/number", 3)
  if type(parent) ~= "table" then return end
  local exclude = parent._exclude
  return exclude and exclude[childName]
end
function Configuration:Overriden(parent, childName)
  assertType(parent,"table", childName,"string/number", 3)
  local child = parent[childName]
  if child == nil then return false end
  return child.type ~= nil
end
function Configuration:ConfGet(info)
  local val = self:Get(info)

  if type(val) == "table" and val.r and val.g and val.b then return colUnpack(val) end
  return val
end
function Configuration:ConfGetString(info)
  return tostring(self:ConfGet(info))
end
function Configuration:ConfSet(info, r, g, b, a)
  if b then return self:Set(info, {r = r, g = g, b = b, a = a })
  elseif tonumber(r) then return self:Set(info, tonumber(r))
  else return self:Set(info, r) end
end
function Configuration:GetParent(info)
  local parent = self:Get(K(info):r(1))
  if type(parent) == "table" then return parent end
end
function Configuration:GetParentName(info)
  return info[#info - 1]
end
function Configuration:ConfGetAll(info)
  local parent = self:GetParent(info)
  if parent then return self:GetAll(parent, info[#info]) end
end
function Configuration:ConfDelete(info)
  local parent = self:GetParent(info)
  if parent then return self:Delete(parent, info[#info]) end
end
function Configuration:ConfExcluded(info)
  local parent = self:GetParent(info)
  if parent then return self:Excluded(parent, info[#info]) end
end
function Configuration:ConfOverriden(info)
  local parent = self:GetParent(info)
  if parent then return self:Overriden(parent, info[#info]) end
end
function Configuration:Number(info, val)
  return val and type(tonumber(val)) == "number"
end
function Configuration:Locked(info)
  local parent = self:GetParent(info)
  if parent then return parent.lock end
end

D.Configuration = Configuration


-----------------
-- Addon handler
-----------------

local function exportFunc(parent, funcName)
  return function(_, ...)
    return parent[funcName](parent, ...)
  end
end
D.exportFunc = exportFunc

local Handler = Configuration:New()

function Handler:New(addon)
  assertType(addon,"table", 3)
  local handler = Handler._super.New(self)
  setmetatable(addon, Handler)
  local addon_settings = addon.settings
  if addon_settings then
    addon.defaults = addon_settings.defaults
    addon_settings.handler = addon
  end
  addon.timers = {}
  addon.frames = {}
  return addon
end
function D.Addon(shortName, name, ...)
  local addon = LibStub("AceAddon-3.0"):NewAddon(shortName, "AceConsole-3.0", "AceEvent-3.0", "AceTimer-3.0", ...)
  addon.shortName = shortName
  addon.name = name
  addon.mediaPath = "Interface\\AddOns\\" .. shortName .. "\\Media\\"
  addon.settings = LibStub(shortName .. "-Settings")
  addon.exports = LibStub:NewLibrary(shortName, 1)
  return Handler:New(addon)
end
function Handler:Output(...)
  if self.core._debug and self.core._debug.print then print(self.name," | ",say(...)) end
end
function Handler:GetVersion()
  return self.core._version
end
function Handler:SetVersion(version)
  self.core._version = version
end
function Handler:ResetDB()
  self.lib.db:ResetProfile()
end
function Handler:Export(...)
  local exports = self.exports
  for _,funcName in pairs{ ... } do
    exports[funcName] = exportFunc(self, funcName)
  end
end
function Handler:Queue(funcName, ...)
  local timers = self.timers
  if type(self[funcName]) ~= "function" then error(say("Invalid function ", funcName,"."), 2) end
  if not self.ScheduleTimer then error("AceTimer not mixed in!", 2) end
  if not timers[funcName] or self:TimeLeft(timers[funcName]) == 0 then
    timers[funcName] = self:ScheduleTimer(funcName,0.01,...)
  end
end
function Handler:TrackEvent(funcName, event, ...)
  debugInfo = { funcName, event, " -> ", tconcat({...}, ", ") }
  event = event or funcName
  self:RegisterEvent(event, funcName, ...)
end
function Handler:TrackEventQueue(funcName, event, ...)
  event = event or funcName
  self:RegisterEvent(event, "Queue", funcName, ...)
end
function Handler:TrackMultiEvent(funcName, ...)
  self:Map("TrackEvent", { funcName }, {...})
end
function Handler:TrackMultiEventQueue(funcName, ...)
  self:Map("TrackEventQueue", { funcName }, {...})
end
function Handler:Reset(registered)
  self.core = self.lib.db.profile
  self:CrawlDefaults()
  if registered ~= false then
    self:SetVersion(self.version)
  end
  if self:ConfGetDebug{ "export" } then
    _G[self.shortName .. "Debug"] = self
  else
    _G[self.shortName .. "Debug"] = nil
  end

  local settings = self.settings
  if settings then
    settings.handler = self
    settings.options = { type = "group", args = {} }
    local defaults = self.defaults
    if defaults and defaults.options then
      settings.options = defaults.options
    end
    settings.options.handler = self
    settings.crawler = NodeCrawler:New(settings.options.args)

    if defaults and defaults.global then
      setIndex(self.core, defaults.global)
    end
  end
  self:OnReset(registered)
  if self.OnLoad then self:OnLoad(registered) end

  if settings and settings.options then
    self.lib.registry:RegisterOptionsTable(self.name, settings.options)
  end
end

function Handler:MigrateDB(...)
  local keys = { ... }
  local old = Configuration:New{}
  for i,key in pairs(keys) do
    key = K(key)
    if key == nil then key = {} end
    local subTable = self:Get(key)
    if type(subTable) ~= "table"then
      old:Set(key, subTable)
    else
      for k,v in pairs(subTable) do
        local subKey = key:a(k)
        if type(v) ~= "table" then old:Set(subKey, v) end
      end
    end
  end

  self:ResetDB()

  for i,key in pairs(keys) do
    key = K(key)
    local subTable = old:Get(key)
    if type(subTable) ~= "table" then
      if subTable ~= self:ConfGet(key) then self:ConfSet(key, subTable) end
    else for k,v in pairs(subTable) do
      local subKey = key:a(k)
      if type(v) ~= "table" and v ~= self:ConfGet(subKey) then self:ConfSet(subKey, v) end
    end end
  end

  if self.version then self:SetVersion(self.version) end
end

local function debugKey(info)
  return K{ "_debug" }:a(unpack(info))
end
function Handler:ConfGetDebug(info)
  return self:ConfGet(debugKey(info))
end
function Handler:ConfSetDebug(info, r, g, b, a)
  self:ConfSet(debugKey(info), r, g, b, a)
end
function Handler:DebugExport(info, val)
  if val then
    _G[self.shortName .. "Debug"] = self
  else
    _G[self.shortName .. "Debug"] = nil
  end
  return self:ConfSetDebug(info,val)
end
function Handler:DebugOptions(args)
  local debug = {
    type = "group",
    handler = self,
    get = "ConfGetDebug",
    set = "ConfSetDebug",
    args = {
      export = {
        type = "toggle",
        name = "Export to global variable (" .. self.shortName .. "Debug)",
        width = "full",
        set = "DebugExport",
        order = 0
      },
      print = {
        type = "toggle",
        name = "Print debug messages",
        width = "full",
        set = "ConfSetDebug",
        order = 1
      },
    }
  }
  local debug_args = debug.args
  for k,v in pairs(args) do
    debug_args[k] = v
  end
  return debug
end

local dbCallbacks = { "OnProfileCopied", "OnProfileDeleted", "OnProfileReset", "OnDatabaseReset", "OnProfileChanged" }
function Handler:Register(libs,defaultProfile,defaults)
  self.lib = self.lib or {}
  local name, lib, settings = self.name, self.lib, self.settings
  if settings and settings.custom then
    settings.defaults = settings.defaults or {}
    local subDefaults = settings.defaults
    local battletag = select(2, BNGetInfo())
    for name, custom in pairs(settings.custom) do
      if name == battletag then
        for customType, customSettings in pairs(custom) do
          subDefaults[customType] = subDefaults[customType] or {}
          local subDefaultType = subDefaults[customType]
          for propertyName, propertyVal in pairs(customSettings) do
            subDefaultType[propertyName] = propertyVal
          end
        end
      end
    end
    if not defaults and subDefaults.db then
      defaults = settings.defaults.db
    end
  end
  defaults = defaults or {}

  lib.db = LibStub("AceDB-3.0"):New(self.shortName .. "DB", defaults, defaultProfile)
  lib.registry = LibStub("AceConfigRegistry-3.0")
  lib.settings = LibStub("AceConfigDialog-3.0")
  lib.options = LibStub("AceDBOptions-3.0")
  for libKey, libName in pairs(libs or {}) do
    if not lib[libKey] then
      lib[libKey] = LibStub(libName,true)
    end
  end

  self:Reset(false)

  lib.settings:AddToBlizOptions(name)
  local profileSettings = LibStub("AceDBOptions-3.0"):GetOptionsTable(lib.db)
  lib.registry:RegisterOptionsTable(name .. " Profile", profileSettings)
  lib.settings:AddToBlizOptions(name .. " Profile", "Profiles", name)

  for _,callback in pairs(dbCallbacks) do
    lib.db.RegisterCallback(self, callback, "Reset")
  end

  if settings and settings.debug then
    local debug = self:DebugOptions(settings.debug)
    lib.registry:RegisterOptionsTable(name .. " Extras", debug)
    lib.settings:AddToBlizOptions(name .. " Extras","Extras", name)
  end
end

function Handler:Draggable(frame, conf, funcName)
  if frame:GetScript("OnReceiveDrag") then return end
  funcName = funcName or "Rebuild"
  local func = function()
    if type(funcName) == "function" then funcName()
    elseif type(self[funcName]) == "function" then self[funcName](self) end
    self.lib.registry:NotifyChange(self.name)
  end
  draggable(frame, conf, func)
end

D.Handler = Handler